@@ -2,7 +2,7 @@ const NEWLINE = /\r?\n/;
 
 module.exports = grammar({
   name: "gleam",
-  externals: ($) => [$.quoted_content],
+  externals: ($) => [$.quoted_content, $.doc_comment_content],
   extras: ($) => [
     NEWLINE,
     /\s/,
@@ -40,9 +40,18 @@ module.exports = grammar({
       ),
 
     /* Comments */
-    module_comment: ($) => token(seq("////", /.*/)),
-    statement_comment: ($) => token(seq("///", /.*/)),
-    comment: ($) => token(seq("//", /.*/)),
+
+    module_comment: ($) => prec(3, seq("////", $.doc_comment_content)),
+
+    statement_comment: ($) => prec(2, seq("///", $.doc_comment_content)),
+
+    comment: ($) =>
+      choice(
+        prec(1, seq("//", /.*/)),
+
+        // `///// ...` looks like a `module_comment`, but it is not
+        prec(4, seq("/////", /.*/))
+      ),
 
     /* Target groups
      * DEPRECATED: This syntax was replaced with attributes in v0.30.
@@ -62,7 +71,7 @@ module.exports = grammar({
       seq(
         "@",
         field("name", $.identifier),
-        field("arguments", alias($._attribute_arguments, $.arguments))
+        optional(field("arguments", alias($._attribute_arguments, $.arguments)))
       ),
 
     _attribute_arguments: ($) =>
@@ -92,9 +101,13 @@ module.exports = grammar({
           optional(seq("as", field("alias", $.identifier)))
         ),
         seq(
-          optional("type"),
+          "type",
           field("name", $.type_identifier),
           optional(seq("as", field("alias", $.type_identifier)))
+        ),
+        seq(
+          field("name", $.constructor_name),
+          optional(seq("as", field("alias", $.constructor_name)))
         )
       ),
 
@@ -118,7 +131,8 @@ module.exports = grammar({
         alias($._constant_bit_string, $.bit_string),
         alias($.constant_record, $.record),
         $.identifier,
-        alias($.constant_field_access, $.field_access)
+        alias($.constant_field_access, $.field_access),
+        alias($.constant_binary_expression, $.binary_expression)
       ),
     constant_tuple: ($) =>
       seq("#", "(", optional(series_of($._constant_value, ",")), ")"),
@@ -139,10 +153,15 @@ module.exports = grammar({
         ")"
       ),
     constant_record_argument: ($) =>
-      seq(
-        optional(seq(field("label", $.label), ":")),
-        field("value", $._constant_value)
+      choice(
+        seq(
+          optional(seq(field("label", $.label), ":")),
+          field("value", $._constant_value)
+        ),
+        seq(field("label", $.label), ":")
       ),
+    constant_binary_expression: ($) =>
+      choice(binaryExpr(prec.left, 7, "<>", $._constant_value)),
     // This rule exists to parse remote function references which are generally
     // indistinguishable from field accesses and so share an AST node.
     constant_field_access: ($) =>
@@ -251,13 +270,7 @@ module.exports = grammar({
           field("name", $.identifier),
           field("parameters", $.function_parameters),
           optional(seq("->", field("return_type", $._type))),
-          optional(
-            seq(
-              "{",
-              optional(field("body", alias($._statement_seq, $.function_body))),
-              "}"
-            )
-          )
+          optional(field("body", $.block))
         )
       ),
     function_parameters: ($) =>
@@ -279,7 +292,8 @@ module.exports = grammar({
       seq(field("label", $.label), field("name", $.identifier)),
     _name_param: ($) => field("name", $.identifier),
     _statement_seq: ($) => repeat1($._statement),
-    _statement: ($) => choice($._expression, $.let, $.let_assert, $.use),
+    _statement: ($) =>
+      choice($._expression, $.let, $.let_assert, $.use, $.assert),
     _expression: ($) => choice($._expression_unit, $.binary_expression),
     binary_expression: ($) =>
       choice(
@@ -295,7 +309,13 @@ module.exports = grammar({
         binaryExpr(prec.left, 4, ">=", $._expression),
         binaryExpr(prec.left, 4, ">.", $._expression),
         binaryExpr(prec.left, 4, ">=.", $._expression),
-        binaryExpr(prec.left, 5, "|>", $._expression),
+        binaryExpr(
+          prec.left,
+          5,
+          "|>",
+          $._expression,
+          choice($.pipeline_echo, $._expression)
+        ),
         binaryExpr(prec.left, 6, "+", $._expression),
         binaryExpr(prec.left, 6, "+.", $._expression),
         binaryExpr(prec.left, 6, "-", $._expression),
@@ -327,12 +347,12 @@ module.exports = grammar({
         $.todo,
         $.panic,
         $.tuple,
+        $.echo,
         $.list,
         alias($._expression_bit_string, $.bit_string),
         $.anonymous_function,
         $.block,
         $.case,
-        $.assert,
         $.boolean_negation,
         $.integer_negation,
         $.record_update,
@@ -346,18 +366,44 @@ module.exports = grammar({
         optional(field("arguments", $.arguments))
       ),
     todo: ($) =>
-      seq(
-        "todo",
-        optional(
-          choice(
-            // DEPRECATED: The 'as' syntax was introduced in v0.30.
-            seq("(", field("message", $.string), ")"),
-            seq("as", field("message", $._expression))
+      prec.left(
+        seq(
+          "todo",
+          optional(
+            choice(
+              // DEPRECATED: The 'as' syntax was introduced in v0.30.
+              seq("(", field("message", $.string), ")"),
+              seq("as", field("message", $._expression))
+            )
           )
         )
       ),
     panic: ($) =>
-      seq("panic", optional(seq("as", field("message", $._expression)))),
+      prec.left(
+        seq(
+          "panic",
+          optional(
+            choice(
+              seq("(", field("message", $.string), ")"),
+              seq("as", field("message", $._expression))
+            )
+          )
+        )
+      ),
+
+    pipeline_echo: ($) =>
+      prec.left(
+        seq("echo", optional(seq("as", field("message", $._expression_unit))))
+      ),
+    echo: ($) =>
+      prec.left(
+        seq(
+          "echo",
+          $._expression,
+          optional(seq("as", field("message", $._expression)))
+        )
+      ),
+
     tuple: ($) => seq("#", "(", optional(series_of($._expression, ",")), ")"),
     list: ($) =>
       seq(
@@ -381,9 +427,7 @@ module.exports = grammar({
           alias($.anonymous_function_parameters, $.function_parameters)
         ),
         optional(seq("->", field("return_type", $._type))),
-        "{",
-        field("body", alias($._statement_seq, $.function_body)),
-        "}"
+        field("body", $.block)
       ),
     anonymous_function_parameters: ($) =>
       seq(
@@ -401,7 +445,7 @@ module.exports = grammar({
         choice($._discard_param, $._name_param),
         optional($._type_annotation)
       ),
-    block: ($) => seq("{", $._statement_seq, "}"),
+    block: ($) => seq("{", optional($._statement_seq), "}"),
     case: ($) =>
       seq(
         "case",
@@ -447,7 +491,16 @@ module.exports = grammar({
         binaryExpr(prec.left, 4, ">", $._case_clause_guard_expression),
         binaryExpr(prec.left, 4, ">=", $._case_clause_guard_expression),
         binaryExpr(prec.left, 4, ">.", $._case_clause_guard_expression),
-        binaryExpr(prec.left, 4, ">=.", $._case_clause_guard_expression)
+        binaryExpr(prec.left, 4, ">=.", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 5, "+", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 5, "+.", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 5, "-", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 5, "-.", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 6, "*", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 6, "*.", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 6, "/", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 6, "/.", $._case_clause_guard_expression),
+        binaryExpr(prec.left, 6, "%", $._case_clause_guard_expression)
       ),
     _case_clause_guard_unit: ($) =>
       choice(
@@ -458,7 +511,19 @@ module.exports = grammar({
       ),
     _case_clause_tuple_access: ($) =>
       seq(field("tuple", $.identifier), ".", field("index", $.integer)),
-    let_assert: ($) => seq("let", "assert", $._assignment),
+    let_assert: ($) =>
+      seq(
+        "let",
+        "assert",
+        $._assignment,
+        optional(seq("as", field("message", $._expression)))
+      ),
+    assert: ($) =>
+      seq(
+        "assert",
+        field("value", $._expression),
+        optional(seq("as", field("message", $._expression)))
+      ),
     let: ($) => seq("let", $._assignment),
     use: ($) =>
       seq(
@@ -469,7 +534,6 @@ module.exports = grammar({
       ),
     use_assignments: ($) => series_of($.use_assignment, ","),
     use_assignment: ($) => seq($._pattern, optional($._type_annotation)),
-    assert: ($) => seq("assert", $._assignment),
     boolean_negation: ($) => seq("!", $._expression_unit),
     integer_negation: ($) => seq("-", $._expression_unit),
     _assignment: ($) =>
@@ -494,7 +558,10 @@ module.exports = grammar({
       ),
     record_update_arguments: ($) => series_of($.record_update_argument, ","),
     record_update_argument: ($) =>
-      seq(field("label", $.label), ":", field("value", $._expression)),
+      choice(
+        seq(field("label", $.label), ":", field("value", $._expression)),
+        seq(field("label", $.label), ":")
+      ),
     // As with other AST nodes in this section, `_maybe_record_expression`,
     // `_maybe_tuple_expression`, and `_maybe_function_expresssion` have no
     // corollaries in the Gleam parser. These anonymous AST node denote any
@@ -562,9 +629,12 @@ module.exports = grammar({
     // record arguments, hence the ambiguous name.
     arguments: ($) => seq("(", optional(series_of($.argument, ",")), ")"),
     argument: ($) =>
-      seq(
-        optional(seq(field("label", $.label), ":")),
-        field("value", choice($.hole, $._expression))
+      choice(
+        seq(
+          optional(seq(field("label", $.label), ":")),
+          field("value", choice($.hole, $._expression))
+        ),
+        seq(field("label", $.label), ":")
       ),
     hole: ($) => $._discard_name,
     function_call: ($) =>
@@ -608,9 +678,12 @@ module.exports = grammar({
         ")"
       ),
     record_pattern_argument: ($) =>
-      seq(
-        optional(seq(field("label", $.label), ":")),
-        field("pattern", $._pattern)
+      choice(
+        seq(
+          optional(seq(field("label", $.label), ":")),
+          field("pattern", $._pattern)
+        ),
+        seq(field("label", $.label), ":")
       ),
     pattern_spread: ($) => seq("..", optional(",")),
     tuple_pattern: ($) =>
@@ -652,6 +725,7 @@ module.exports = grammar({
     data_constructors: ($) => repeat1($.data_constructor),
     data_constructor: ($) =>
       seq(
+        optional($.attribute),
         field("name", $.constructor_name),
         optional(field("arguments", $.data_constructor_arguments))
       ),
